@@ -119,8 +119,6 @@ pub struct FormatError {
 pub(crate) enum FormatErrorInner {
     /// Bad framing.
     CrcMismatch {
-        /// bytes to skip to try to recover from this error
-        recover: usize,
         /// Stored CRC32 value
         crc_val: u32,
         /// Calculated CRC32 sum
@@ -198,13 +196,10 @@ pub(crate) enum FormatErrorInner {
     // Errors specific to the IDAT/fDAT chunks.
     /// The compression of the data stream was faulty.
     CorruptFlateStream {
-        err: miniz_oxide::inflate::TINFLStatus,
+        err: fdeflate::DecompressionError,
     },
     /// The image data chunk was too short for the expected pixel count.
     NoMoreImageData,
-    // TODO: strictly type this.
-    /// Filtering of a row has failed.
-    BadFilter(&'static str),
     /// Bad text encoding
     BadTextEncoding(TextDecodingError),
 }
@@ -296,24 +291,8 @@ impl fmt::Display for FormatError {
             NoMoreImageData => write!(fmt, "IDAT or fDAT chunk is has not enough data for image."),
             CorruptFlateStream { err } => {
                 write!(fmt, "Corrupt deflate stream. ")?;
-                use miniz_oxide::inflate::TINFLStatus;
-                match err {
-                    TINFLStatus::Adler32Mismatch => write!(fmt, "Adler32 checksum failed."),
-                    TINFLStatus::BadParam => write!(fmt, "Invalid input parameters."),
-                    // The Done status should already have been handled.
-                    TINFLStatus::Done => write!(fmt, "Unexpected done status."),
-                    TINFLStatus::FailedCannotMakeProgress => {
-                        write!(fmt, "Unexpected end of data.")
-                    }
-                    // The HasMoreOutput status should already have been handled.
-                    TINFLStatus::HasMoreOutput => write!(fmt, "Has more output."),
-                    // The HasMoreInput status should already have been handled.
-                    TINFLStatus::NeedsMoreInput => write!(fmt, "Needs more input."),
-                    // Write out the error number in case a new has been added.
-                    _ => write!(fmt, "Error number {:?}.", err),
-                }
+                write!(fmt, "{:?}", err)
             }
-            BadFilter(message) => write!(fmt, "{}.", message),
             // TODO: Wrap more info in the enum variant
             BadTextEncoding(tde) => {
                 match tde {
@@ -382,19 +361,59 @@ impl From<TextDecodingError> for DecodingError {
     }
 }
 
-/// Decode config
-#[derive(Default)]
-pub(crate) struct DecodeConfig {
+/// Decoder configuration options
+#[derive(Clone)]
+pub struct DecodeOptions {
+    ignore_adler32: bool,
+    ignore_crc: bool,
     ignore_text_chunk: bool,
 }
 
-impl DecodeConfig {
+impl Default for DecodeOptions {
+    fn default() -> Self {
+        Self {
+            ignore_adler32: true,
+            ignore_crc: false,
+            ignore_text_chunk: false,
+        }
+    }
+}
+
+impl DecodeOptions {
+    /// When set, the decoder will not compute and verify the Adler-32 checksum.
+    ///
+    /// Defaults to `true`.
+    pub fn set_ignore_adler32(&mut self, ignore_adler32: bool) {
+        self.ignore_adler32 = ignore_adler32;
+    }
+
+    /// When set, the decoder will not compute and verify the CRC code.
+    ///
+    /// Defaults to `false`.
+    pub fn set_ignore_crc(&mut self, ignore_crc: bool) {
+        self.ignore_crc = ignore_crc;
+    }
+
+    /// Flag to ignore computing and verifying the Adler-32 checksum and CRC
+    /// code.
+    pub fn set_ignore_checksums(&mut self, ignore_checksums: bool) {
+        self.ignore_adler32 = ignore_checksums;
+        self.ignore_crc = ignore_checksums;
+    }
+
+    /// Ignore text chunks while decoding.
+    ///
+    /// Defaults to `false`.
     pub fn set_ignore_text_chunk(&mut self, ignore_text_chunk: bool) {
         self.ignore_text_chunk = ignore_text_chunk;
     }
 }
 
 /// PNG StreamingDecoder (low-level interface)
+///
+/// By default, the decoder does not verify Adler-32 checksum computation. To
+/// enable checksum verification, set it with [`StreamingDecoder::set_ignore_adler32`]
+/// before starting decompression.
 pub struct StreamingDecoder {
     state: Option<State>,
     current_chunk: ChunkState,
@@ -407,7 +426,7 @@ pub struct StreamingDecoder {
     /// Stores where in decoding an `fdAT` chunk we are.
     apng_seq_handled: bool,
     have_idat: bool,
-    decode_config: DecodeConfig,
+    decode_options: DecodeOptions,
 }
 
 struct ChunkState {
@@ -430,15 +449,22 @@ impl StreamingDecoder {
     ///
     /// Allocates the internal buffers.
     pub fn new() -> StreamingDecoder {
+        StreamingDecoder::new_with_options(DecodeOptions::default())
+    }
+
+    pub fn new_with_options(decode_options: DecodeOptions) -> StreamingDecoder {
+        let mut inflater = ZlibStream::new();
+        inflater.set_ignore_adler32(decode_options.ignore_adler32);
+
         StreamingDecoder {
             state: Some(State::Signature(0, [0; 7])),
             current_chunk: ChunkState::default(),
-            inflater: ZlibStream::new(),
+            inflater,
             info: None,
             current_seq_no: None,
             apng_seq_handled: false,
             have_idat: false,
-            decode_config: DecodeConfig::default(),
+            decode_options,
         }
     }
 
@@ -448,7 +474,7 @@ impl StreamingDecoder {
         self.current_chunk.crc = Crc32::new();
         self.current_chunk.remaining = 0;
         self.current_chunk.raw_bytes.clear();
-        self.inflater = ZlibStream::new();
+        self.inflater.reset();
         self.info = None;
         self.current_seq_no = None;
         self.apng_seq_handled = false;
@@ -460,13 +486,32 @@ impl StreamingDecoder {
         self.info.as_ref()
     }
 
-    /// Set decode config
-    pub(crate) fn set_decode_config(&mut self, decode_config: DecodeConfig) {
-        self.decode_config = decode_config;
+    pub fn set_ignore_text_chunk(&mut self, ignore_text_chunk: bool) {
+        self.decode_options.set_ignore_text_chunk(ignore_text_chunk);
     }
 
-    pub fn set_ignore_text_chunk(&mut self, ignore_text_chunk: bool) {
-        self.decode_config.set_ignore_text_chunk(ignore_text_chunk);
+    /// Return whether the decoder is set to ignore the Adler-32 checksum.
+    pub fn ignore_adler32(&self) -> bool {
+        self.inflater.ignore_adler32()
+    }
+
+    /// Set whether to compute and verify the Adler-32 checksum during
+    /// decompression. Return `true` if the flag was successfully set.
+    ///
+    /// The decoder defaults to `true`.
+    ///
+    /// This flag cannot be modified after decompression has started until the
+    /// [`StreamingDecoder`] is reset.
+    pub fn set_ignore_adler32(&mut self, ignore_adler32: bool) -> bool {
+        self.inflater.set_ignore_adler32(ignore_adler32)
+    }
+
+    /// Set whether to compute and verify the Adler-32 checksum during
+    /// decompression. Return `true` if the flag was successfully set.
+    ///
+    /// The decoder defaults to `false`.
+    pub fn set_ignore_crc(&mut self, ignore_crc: bool) {
+        self.decode_options.set_ignore_crc(ignore_crc)
     }
 
     /// Low level StreamingDecoder interface.
@@ -500,25 +545,6 @@ impl StreamingDecoder {
     ) -> Result<(usize, Decoded), DecodingError> {
         use self::State::*;
 
-        macro_rules! goto (
-            ($n:expr, $state:expr) => ({
-                self.state = Some($state);
-                Ok(($n, Decoded::Nothing))
-            });
-            ($state:expr) => ({
-                self.state = Some($state);
-                Ok((1, Decoded::Nothing))
-            });
-            ($n:expr, $state:expr, emit $res:expr) => ({
-                self.state = Some($state);
-                Ok(($n, $res))
-            });
-            ($state:expr, emit $res:expr) => ({
-                self.state = Some($state);
-                Ok((1, $res))
-            })
-        );
-
         let current_byte = buf[0];
 
         // Driver should ensure that state is never None
@@ -527,12 +553,14 @@ impl StreamingDecoder {
         match state {
             Signature(i, mut signature) if i < 7 => {
                 signature[i as usize] = current_byte;
-                goto!(Signature(i + 1, signature))
+                self.state = Some(Signature(i + 1, signature));
+                Ok((1, Decoded::Nothing))
             }
             Signature(_, signature)
                 if signature == [137, 80, 78, 71, 13, 10, 26] && current_byte == 10 =>
             {
-                goto!(U32(U32Value::Length))
+                self.state = Some(U32(U32Value::Length));
+                Ok((1, Decoded::Nothing))
             }
             Signature(..) => Err(DecodingError::Format(
                 FormatErrorInner::InvalidSignature.into(),
@@ -541,7 +569,10 @@ impl StreamingDecoder {
                 use self::U32Value::*;
                 val |= u32::from(current_byte);
                 match type_ {
-                    Length => goto!(U32(Type(val))),
+                    Length => {
+                        self.state = Some(U32(Type(val)));
+                        Ok((1, Decoded::Nothing))
+                    }
                     Type(length) => {
                         let type_str = ChunkType([
                             (val >> 24) as u8,
@@ -556,38 +587,40 @@ impl StreamingDecoder {
                             self.current_chunk.type_ = type_str;
                             self.inflater.finish_compressed_chunks(image_data)?;
                             self.inflater.reset();
-                            return goto!(
-                                0,
-                                U32Byte3(Type(length), val & !0xff),
-                                emit Decoded::ImageDataFlushed
-                            );
+                            self.state = Some(U32Byte3(Type(length), val & !0xff));
+                            return Ok((0, Decoded::ImageDataFlushed));
                         }
                         self.current_chunk.type_ = type_str;
-                        self.current_chunk.crc.reset();
-                        self.current_chunk.crc.update(&type_str.0);
+                        if !self.decode_options.ignore_crc {
+                            self.current_chunk.crc.reset();
+                            self.current_chunk.crc.update(&type_str.0);
+                        }
                         self.current_chunk.remaining = length;
                         self.apng_seq_handled = false;
                         self.current_chunk.raw_bytes.clear();
-                        goto!(
-                            ReadChunk(type_str),
-                            emit Decoded::ChunkBegin(length, type_str)
-                        )
+                        self.state = Some(ReadChunk(type_str));
+                        Ok((1, Decoded::ChunkBegin(length, type_str)))
                     }
                     Crc(type_str) => {
-                        let sum = self.current_chunk.crc.clone().finalize();
-                        if CHECKSUM_DISABLED || val == sum {
-                            goto!(
-                                State::U32(U32Value::Length),
-                                emit if type_str == IEND {
-                                    Decoded::ImageEnd
-                                } else {
-                                    Decoded::ChunkComplete(val, type_str)
-                                }
-                            )
+                        // If ignore_crc is set, do not calculate CRC. We set
+                        // sum=val so that it short-circuits to true in the next
+                        // if-statement block
+                        let sum = if self.decode_options.ignore_crc {
+                            val
+                        } else {
+                            self.current_chunk.crc.clone().finalize()
+                        };
+
+                        if val == sum || CHECKSUM_DISABLED {
+                            self.state = Some(State::U32(U32Value::Length));
+                            if type_str == IEND {
+                                Ok((1, Decoded::ImageEnd))
+                            } else {
+                                Ok((1, Decoded::ChunkComplete(val, type_str)))
+                            }
                         } else {
                             Err(DecodingError::Format(
                                 FormatErrorInner::CrcMismatch {
-                                    recover: 1,
                                     crc_val: val,
                                     crc_sum: sum,
                                     chunk: type_str,
@@ -598,18 +631,24 @@ impl StreamingDecoder {
                     }
                 }
             }
-            U32Byte2(type_, val) => goto!(U32Byte3(type_, val | u32::from(current_byte) << 8)),
-            U32Byte1(type_, val) => goto!(U32Byte2(type_, val | u32::from(current_byte) << 16)),
-            U32(type_) => goto!(U32Byte1(type_, u32::from(current_byte) << 24)),
+            U32Byte2(type_, val) => {
+                self.state = Some(U32Byte3(type_, val | u32::from(current_byte) << 8));
+                Ok((1, Decoded::Nothing))
+            }
+            U32Byte1(type_, val) => {
+                self.state = Some(U32Byte2(type_, val | u32::from(current_byte) << 16));
+                Ok((1, Decoded::Nothing))
+            }
+            U32(type_) => {
+                self.state = Some(U32Byte1(type_, u32::from(current_byte) << 24));
+                Ok((1, Decoded::Nothing))
+            }
             PartialChunk(type_str) => {
                 match type_str {
                     IDAT => {
                         self.have_idat = true;
-                        goto!(
-                            0,
-                            DecodeData(type_str, 0),
-                            emit Decoded::PartialChunk(type_str)
-                        )
+                        self.state = Some(DecodeData(type_str, 0));
+                        Ok((0, Decoded::PartialChunk(type_str)))
                     }
                     chunk::fdAT => {
                         let data_start;
@@ -637,11 +676,8 @@ impl StreamingDecoder {
                                 FormatErrorInner::MissingFctl.into(),
                             ));
                         }
-                        goto!(
-                            0,
-                            DecodeData(type_str, data_start),
-                            emit Decoded::PartialChunk(type_str)
-                        )
+                        self.state = Some(DecodeData(type_str, data_start));
+                        Ok((0, Decoded::PartialChunk(type_str)))
                     }
                     // Handle other chunks
                     _ => {
@@ -653,10 +689,8 @@ impl StreamingDecoder {
                             // We need it fully before parsing.
                             self.reserve_current_chunk()?;
 
-                            goto!(
-                                0, ReadChunk(type_str),
-                                emit Decoded::PartialChunk(type_str)
-                            )
+                            self.state = Some(ReadChunk(type_str));
+                            Ok((0, Decoded::PartialChunk(type_str)))
                         }
                     }
                 }
@@ -665,7 +699,8 @@ impl StreamingDecoder {
                 // The _previous_ event wanted to return the contents of raw_bytes, and let the
                 // caller consume it,
                 if self.current_chunk.remaining == 0 {
-                    goto!(0, U32(U32Value::Crc(type_str)))
+                    self.state = Some(U32(U32Value::Crc(type_str)));
+                    Ok((0, Decoded::Nothing))
                 } else {
                     let ChunkState {
                         crc,
@@ -678,7 +713,8 @@ impl StreamingDecoder {
                     let bytes_avail = min(buf.len(), buf_avail);
                     let n = min(*remaining, bytes_avail as u32);
                     if buf_avail == 0 {
-                        goto!(0, PartialChunk(type_str))
+                        self.state = Some(PartialChunk(type_str));
+                        Ok((0, Decoded::Nothing))
                     } else {
                         let buf = &buf[..n as usize];
                         crc.update(buf);
@@ -686,10 +722,11 @@ impl StreamingDecoder {
 
                         *remaining -= n;
                         if *remaining == 0 {
-                            goto!(n as usize, PartialChunk(type_str))
+                            self.state = Some(PartialChunk(type_str));
                         } else {
-                            goto!(n as usize, ReadChunk(type_str))
+                            self.state = Some(ReadChunk(type_str));
                         }
+                        Ok((n as usize, Decoded::Nothing))
                     }
                 }
             }
@@ -700,17 +737,11 @@ impl StreamingDecoder {
                 n += c;
                 if n == chunk_len && c == 0 {
                     self.current_chunk.raw_bytes.clear();
-                    goto!(
-                        0,
-                        ReadChunk(type_str),
-                        emit Decoded::ImageData
-                    )
+                    self.state = Some(ReadChunk(type_str));
+                    Ok((0, Decoded::ImageData))
                 } else {
-                    goto!(
-                        0,
-                        DecodeData(type_str, n),
-                        emit Decoded::ImageData
-                    )
+                    self.state = Some(DecodeData(type_str, n));
+                    Ok((0, Decoded::ImageData))
                 }
             }
         }
@@ -719,7 +750,7 @@ impl StreamingDecoder {
     fn reserve_current_chunk(&mut self) -> Result<(), DecodingError> {
         // FIXME: use limits, also do so in iccp/zlib decompression.
         const MAX: usize = 0x10_0000;
-        let ref mut buffer = self.current_chunk.raw_bytes;
+        let buffer = &mut self.current_chunk.raw_bytes;
 
         // Double if necessary, but no more than until the limit is reached.
         let reserve_size = MAX.saturating_sub(buffer.capacity()).min(buffer.len());
@@ -750,9 +781,9 @@ impl StreamingDecoder {
             chunk::cHRM => self.parse_chrm(),
             chunk::sRGB => self.parse_srgb(),
             chunk::iCCP => self.parse_iccp(),
-            chunk::tEXt if !self.decode_config.ignore_text_chunk => self.parse_text(),
-            chunk::zTXt if !self.decode_config.ignore_text_chunk => self.parse_ztxt(),
-            chunk::iTXt if !self.decode_config.ignore_text_chunk => self.parse_itxt(),
+            chunk::tEXt if !self.decode_options.ignore_text_chunk => self.parse_text(),
+            chunk::zTXt if !self.decode_options.ignore_text_chunk => self.parse_ztxt(),
+            chunk::iTXt if !self.decode_options.ignore_text_chunk => self.parse_itxt(),
             _ => Ok(Decoded::PartialChunk(type_str)),
         } {
             Err(err) => {
@@ -792,7 +823,7 @@ impl StreamingDecoder {
             }
             0
         });
-        self.inflater = ZlibStream::new();
+        self.inflater.reset();
         let fc = FrameControl {
             sequence_number: next_seq_no,
             width: buf.read_be()?,
@@ -849,9 +880,9 @@ impl StreamingDecoder {
         let info = self.info.as_mut().unwrap();
         if info.palette.is_some() {
             // Only one palette is allowed
-            return Err(DecodingError::Format(
+            Err(DecodingError::Format(
                 FormatErrorInner::DuplicateChunk { kind: chunk::PLTE }.into(),
-            ));
+            ))
         } else {
             info.palette = Some(Cow::Owned(self.current_chunk.raw_bytes.clone()));
             Ok(Decoded::Nothing)
@@ -926,9 +957,9 @@ impl StreamingDecoder {
                 FormatErrorInner::AfterIdat { kind: chunk::pHYs }.into(),
             ))
         } else if info.pixel_dims.is_some() {
-            return Err(DecodingError::Format(
+            Err(DecodingError::Format(
                 FormatErrorInner::DuplicateChunk { kind: chunk::pHYs }.into(),
-            ));
+            ))
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
             let xppu = buf.read_be()?;
@@ -954,13 +985,10 @@ impl StreamingDecoder {
             Err(DecodingError::Format(
                 FormatErrorInner::AfterIdat { kind: chunk::cHRM }.into(),
             ))
-        } else if info.source_chromaticities.is_some() {
-            return Err(DecodingError::Format(
+        } else if info.chrm_chunk.is_some() {
+            Err(DecodingError::Format(
                 FormatErrorInner::DuplicateChunk { kind: chunk::cHRM }.into(),
-            ));
-        } else if info.srgb.is_some() {
-            // Ignore chromaticities if sRGB profile is used.
-            Ok(Decoded::Nothing)
+            ))
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
             let white_x: u32 = buf.read_be()?;
@@ -991,7 +1019,12 @@ impl StreamingDecoder {
                 ),
             };
 
-            info.source_chromaticities = Some(source_chromaticities);
+            info.chrm_chunk = Some(source_chromaticities);
+            // Ignore chromaticities if sRGB profile is used.
+            if info.srgb.is_none() {
+                info.source_chromaticities = Some(source_chromaticities);
+            }
+
             Ok(Decoded::Nothing)
         }
     }
@@ -1002,17 +1035,21 @@ impl StreamingDecoder {
             Err(DecodingError::Format(
                 FormatErrorInner::AfterIdat { kind: chunk::gAMA }.into(),
             ))
-        } else if info.source_gamma.is_some() {
-            return Err(DecodingError::Format(
+        } else if info.gama_chunk.is_some() {
+            Err(DecodingError::Format(
                 FormatErrorInner::DuplicateChunk { kind: chunk::gAMA }.into(),
-            ));
-        } else if info.srgb.is_some() {
-            // Ignore chromaticities if sRGB profile is used.
-            Ok(Decoded::Nothing)
+            ))
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
             let source_gamma: u32 = buf.read_be()?;
-            self.info.as_mut().unwrap().source_gamma = Some(ScaledFloat::from_scaled(source_gamma));
+            let source_gamma = ScaledFloat::from_scaled(source_gamma);
+
+            info.gama_chunk = Some(source_gamma);
+            // Ignore chromaticities if sRGB profile is used.
+            if info.srgb.is_none() {
+                info.source_gamma = Some(source_gamma);
+            }
+
             Ok(Decoded::Nothing)
         }
     }
@@ -1024,9 +1061,9 @@ impl StreamingDecoder {
                 FormatErrorInner::AfterIdat { kind: chunk::acTL }.into(),
             ))
         } else if info.srgb.is_some() {
-            return Err(DecodingError::Format(
+            Err(DecodingError::Format(
                 FormatErrorInner::DuplicateChunk { kind: chunk::sRGB }.into(),
-            ));
+            ))
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
             let raw: u8 = buf.read_be()?; // BE is is nonsense for single bytes, but this way the size is checked.
@@ -1049,9 +1086,9 @@ impl StreamingDecoder {
                 FormatErrorInner::AfterIdat { kind: chunk::iCCP }.into(),
             ))
         } else if info.icc_profile.is_some() {
-            return Err(DecodingError::Format(
+            Err(DecodingError::Format(
                 FormatErrorInner::DuplicateChunk { kind: chunk::iCCP }.into(),
-            ));
+            ))
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
 
@@ -1204,7 +1241,7 @@ impl StreamingDecoder {
         let (keyword_slice, value_slice) = Self::split_keyword(buf)?;
 
         let compression_method = *value_slice
-            .get(0)
+            .first()
             .ok_or_else(|| DecodingError::from(TextDecodingError::InvalidCompressionMethod))?;
 
         let text_slice = &value_slice[1..];
@@ -1223,7 +1260,7 @@ impl StreamingDecoder {
         let (keyword_slice, value_slice) = Self::split_keyword(buf)?;
 
         let compression_flag = *value_slice
-            .get(0)
+            .first()
             .ok_or_else(|| DecodingError::from(TextDecodingError::MissingCompressionFlag))?;
 
         let compression_method = *value_slice
